# Vivado HLS run script
#
# Usage:
#        Simulation: vivado_hls -f vivado_hls.tcl sim
#      RTL Generate: vivado_hls -f vivado_hls.tcl rtl

set workspace     "HLS_workspace"
set top_level     "hls_ciphercore"
set target_chip   "xc7vx690tffg1761-2"
set target_period 100
set solution      "solution1"

proc setup {} {
    open_project $::workspace
    set_top $::top_level
    add_files { ../src_hls/hls_aes.cpp }
    add_files { ../src_hls/hls_deoxys.cpp }
    add_files { ../src_hls/hls_utils.cpp }
    add_files { ../src_hls/hls_aes.h }
    add_files { ../src_hls/hls_deoxys.h }
    add_files { ../src_hls/hls_interface.h }
    add_files { ../src_hls/hls_utils.h }
    add_files -tb { ../src_hls/hls_tb.cpp }
    open_solution $::solution
    set_part $::target_chip
    create_clock -period 100 -name default
    config_rtl -encoding auto -reset all -reset_level high
}
proc sim {} { csim_design -clean }
proc rtl {} { csynth_design }


proc main {} {
    setup
    foreach arg $::argv {
        if {$arg == "sim"} { sim }
        if {$arg == "rtl"} { rtl }
    }
}

main
exit
