/*
 * @author     Ekawat (ice) Homsirikamol
 * @version    1.0
 * @copyright  Copyright (c) 2017 Cryptographic Engineering Research Group
 *             ECE Department, George Mason University Fairfax, VA, U.S.A.
 *             All rights Reserved.
 * @license    This project is released under the GNU Public License.
 *             The license and distribution terms for this file may be
 *             found in the file LICENSE in this distribution or at
 *             http://www.gnu.org/licenses/gpl-3.0.txt
 * @note       This implementation is derived in part from the reference implementation
 *             included in SUPERCOP-20171020, with the list of co-authors listed in ref/designers
 *             
 *             This is publicly available encryption source code that falls
 *             under the License Exception TSU (Technology and software-
 *             unrestricted)
 */

#include <iostream>
#include <fstream>
#include <string>
#include <stdlib.h> /* strtol */
using namespace std;
#include "hls_interface.h"


data_t str2block(std::string& str)
{
    data_t out = 0;
    unsigned int x;
    //cout << "Parsing: " << str << endl;
    for (int i=0; i<str.length()/2; i++) {
        x = strtoul(str.substr(i*2,2).c_str(), NULL, 16);
        out = (out << 8) ^ x;
    }
    //cout << "[Done]: " << out << endl;
    return out;
}

size_t get_instream(ifstream& fdi, BDI_STREAM& bdi_stream, KEY_STREAM& key_stream, bool* op, bool* key_update) 
{
    char *pEnd;
    bool eoi, eot, partial;
    int  size, type;
    string line;
    string data;
    string keydata;
    string s;
    ap_uint<BLOCKSIZE/8+1> bcalc;
    ap_uint<BLOCKSIZE/8+1> pcalc;
    size_t sz = 0;
    *key_update = 0;

    while(getline(fdi, line)) {
        cout << "IN : " << line << endl;
        if (line.compare("") == 0) continue;
        // Read operation (0 = Encryption, 1 = Decryption)
        if (line.compare("#NEW") == 0) {
            getline(fdi, line); // Message number line
            cout << "IN : " << line << endl;
            getline(fdi, line); // Operation line
            cout << "IN : " << line << endl;
            *op = strtol(line.c_str(), &pEnd, 2);
            continue;
        }
        // Prepare key input 
        if (line.compare("#KEY") == 0) {
            getline(fdi, line);
            cout << "IN : " << line << endl;
            *key_update = strtol(line.c_str(), &pEnd, 2);
            getline(fdi, line);
            cout << "IN : " << line << endl;
            keydata = line;
            // Write data to stream
            if (*key_update == 1) {
                size_t keysize = keydata.length();
                string tmp_data;
                for (size_t i=0; i<(keysize/2)/(KEYBUS/8); i++) {
                    secret_bus tmp;
                    tmp_data = keydata.substr(i*2*(KEYBUS/8),2*(KEYBUS/8));
                    cout << "KEY " << tmp_data.length() << ": " << setw(16) << hex << uppercase << tmp_data << endl;
                    tmp.data = str2block(tmp_data);
                    key_stream.write(tmp);
                }
            }
            continue;
        }
        // End of input data
        if (line.compare("#END") == 0) break;
        // End of file
        if (line.compare("#EOF") == 0) break;
        
        // Read data line
        size_t pos = 0, new_pos;
        // Data
        new_pos = line.find_first_of(',',pos);
        s = line.substr(pos,new_pos-pos);
        data = s;
        pos = new_pos+1; 
        // Size
        new_pos = line.find_first_of(',',pos);
        s = line.substr(pos,new_pos-pos);
        size = strtol(s.c_str(), &pEnd, 10);
        pos = new_pos+1; 
        // Type
        new_pos = line.find_first_of(',',pos);
        s = line.substr(pos,new_pos-pos);
        type = strtol(s.c_str(), &pEnd, 10);
        pos = new_pos+1; 
        // EOI
        new_pos = line.find_first_of(',',pos);
        s = line.substr(pos,new_pos-pos);
        eoi = strtol(s.c_str(), &pEnd, 2);
        pos = new_pos+1; 
        // EOT
        new_pos = line.find_first_of(',',pos);
        s = line.substr(pos,new_pos-pos);
        eot = strtol(s.c_str(), &pEnd, 2);
        pos = new_pos+1; 
        // Partial
        s = line.substr(pos,string::npos);
        partial = strtol(s.c_str(), &pEnd, 2);

        // valid_bytes calculation
        ap_uint<BLOCKSIZE/8+1> t = 1 << BLOCKSIZE/8;
        bcalc = (t - 1) ^ ((t >> size) - 1);
        // pad_loc calculation
        pcalc = 1 << ((BLOCKSIZE/8)-1);
        pcalc = pcalc >> size;

        // Write data to stream
        public_bus tmp;
        tmp.data = str2block(data);
        tmp.valid_bytes = (data_bytes_t) bcalc;
        tmp.pad_loc = (data_bytes_t) pcalc;
        tmp.size = (bsize_t) size;
        tmp.type = (ap_uint<3>) type;
        tmp.eoi = eoi;
        tmp.eot = eot;
        tmp.partial = partial;
        bdi_stream.write(tmp);
        sz++;
    }
    return sz;
}

size_t get_outstream(ifstream& fdi, BDO_STREAM& bdo_stream) 
{
    char *pEnd;
    bool op;
    bool new_key;
    bool eoi, eot, partial;
    int  size, type;
    string line;
    string data;
    string key;
    string s;
    size_t sz = 0;

    while(getline(fdi, line)) {
        cout << "OUT: " << line << endl;
        // Skip empty
        if (line.compare("") == 0) continue;
        // Ignore #NEW 
        if (line.compare("#NEW") == 0) {
            getline(fdi, line); // Read message number line
            continue;
        }
        // End of input data
        if (line.compare("#END") == 0) break;
        // End of file
        if (line.compare("#EOF") == 0) break;
        
        // Read data line
        size_t pos = 0, new_pos;
        // Data
        new_pos = line.find_first_of(',',pos);
        s = line.substr(pos,new_pos-pos);
        data = s;
        pos = new_pos+1; 
        // Size
        new_pos = line.find_first_of(',',pos);
        s = line.substr(pos,new_pos-pos);
        size = strtol(s.c_str(), &pEnd, 10);

        // Write data to stream
        output_bus tmp;
        tmp.data = str2block(data);
        tmp.size = size;
        bdo_stream.write(tmp);
        sz++;
    }
    return sz;
}

void read_comments(ifstream& f)
{
    string line;
    if (f.is_open()) {
        while (getline(f, line))            
            if (line.compare("#BEGIN") == 0) break;
    } else {
        cout << "File not opened" << endl;
    }
}

int main(int argc, char* argv[]) {
    size_t msgno = 0;
    size_t isize, osize;
    ifstream fdi, fdo;

    cout << " =========== \n\n\n";

    /* Read PDI */ 
    fdi.open("../../../../../scripts/cc_di.txt");
    fdo.open("../../../../../scripts/cc_do.txt");
    
    /* Cycle through comments */
    read_comments(fdi);
    read_comments(fdo);
   
    cout << "Starting ..." << endl;

    BDI_STREAM bdi_stream;
    KEY_STREAM key_stream;
    BDO_STREAM bdo_stream;
    BDO_STREAM exp_bdo_stream;
    bool decrypt;
    bool key_update;
    bool msg_auth_valid;
    output_bus exp_bdo;
    output_bus bdo;
    size_t blockno;
    ap_uint<8> out;
    ap_uint<8> exp_out;

    while((isize = get_instream(fdi, bdi_stream, key_stream, &decrypt, &key_update)) > 0) {
        osize = get_outstream(fdo, exp_bdo_stream);
        cout << "Message Number #" << dec << ++msgno 
             << " || Decrypt = " << decrypt
             << " || Total input blocks = " << isize
             << " || Exp total output blocks = " << osize << endl;

        CALL_CRYPTO;  

        
        
        blockno = 1;
        cout << "-- Checking results --" << endl;
        while (osize > 0) {
            exp_bdo = exp_bdo_stream.read();
            bdo = bdo_stream.read();
            cout << "\tExpected: " << setw(16) << hex << uppercase << exp_bdo.data << endl;
            cout << "\tReceived: " << setw(16) << hex << uppercase << bdo.data << endl;
            for (size_t i=0; i<exp_bdo.size; i++) {
                out = (ap_uint<8>) (bdo.data >> (ELEMENTS-1-i)*8); 
                exp_out = (ap_uint<8>) (exp_bdo.data >> (ELEMENTS-1-i)*8); 
                if (out != exp_out) {
                    cout << "\tError! Output mismatched for message number #"
                         << dec << msgno 
                         << " block #" << (int) blockno++ 
                         << " byte #" << (int) (i+1) << endl;
                    cout << hex << uppercase;
                    cout << "\t\tExpected: " << exp_out << endl;
                    cout << "\t\tRecieved: " << out << endl;
                    return 1;
                }
            }
            osize--;
        }
        if (decrypt) {
            cout << "\tmsg_auth_valid = " << msg_auth_valid << endl;
            if (!msg_auth_valid) {
                cout << "\tMessage authentication failed for message number #" << dec << msgno << endl;
                return 1;
            }
        }

    }

    cout << "Congrats! All tests passed." << endl;
    cout << "\n===========\n" << endl;
    /* Close */
    fdi.close();
    fdo.close();
}
